#include "../include/setsublattice.h"

/// <summary>
/// Initializes the (sub)lattice on each node
/// </summary>
void setSubLattice()
{
	/* set standard valeus - assume just one big node */
	thisNode.sites_on_nx = currentConf.data.nx;
	thisNode.sites_on_ny = currentConf.data.ny;
	thisNode.sites_on_nz = currentConf.data.nz;
	thisNode.sites_on_node = currentConf.data.nx * currentConf.data.ny * currentConf.data.nz;
	thisNode.offset_x = 0;
	thisNode.offset_y = 0;
	thisNode.offset_z = 0;
	thisNode.parity = EVEN;
	
	if(thisNode.sites != NULL)
		free(thisNode.sites);
	
#ifdef MPICOMM
	/* closer look - especially with MPI enabled */
	determineDimensions(currentConf.data.nx, currentConf.data.ny, currentConf.data.nz);
	masterPrint("Dimensions per process adjusted");
#endif
	
	/* generate a random configuration to start with or load lattice */
	if(currentConf.data.loadlattice == CREATE_NEW_LATTICE)
		generateRandomConfiguration(thisNode.sites_on_nx, thisNode.sites_on_ny, thisNode.sites_on_nz, currentConf.data.iseed);
	else
		loadLattice(currentConf.data.loadlatticefile);
}

/// <summary>
/// Should be called if no lattice is loaded - we start with a random lattice.
/// Remember to do lots of warmup runs in this case!
/// </summary>
/// <param name="dimx">The thisNode.sites_on_nx number</param>
/// <param name="dimy">The same number for the (sub)-y direction</param>
/// <param name="dimz">An again in z-direction</param>
/// <param name="seed">The specified seed for the random number</param>
void generateRandomConfiguration(int dimx, int dimy, int dimz, int seed)
{
	int k = 0, x, y, z;
	thisNode.sites = (site*)malloc(sizeof(site) * dimx * dimy * dimz);
	
	if(seed == 0)
		srand((unsigned)time(NULL));
	else
		srand((unsigned)seed);
	
	/* loop over all the node's sites */
	for(z = 0; z < dimz; z++)
	{
		for(y = 0; y < dimy; y++)
		{
			for(x = 0; x < dimx; x++)
			{
				thisNode.sites[k].x = x;
				thisNode.sites[k].y = y;
				thisNode.sites[k].z = z;
				thisNode.sites[k].parity = (thisNode.parity + x + y + z) % 2 == 1 ? ODD : EVEN;
				
				/* generate a random value for the sites value */
				thisNode.sites[k].value = (rand() % 3) - 1;
				k++;
			} /* x loop */
		} /* y loop */
	} /* z loop */
	
	masterPrint("Random configuration created");
}

/// <summary>
/// Determines the (sub)lattice dimensions depending on the
/// processor count, overall dimensions and dimension
/// </summary>
/// <param name="nx">The overall NX of the whole lattice</param>
/// <param name="ny">The overall NY of the whole lattice</param>
/// <param name="nz">The overall NZ of the whole lattice</param>
void determineDimensions(int nx, int ny, int nz)
{
	int v, V, offV, i, j, k;
	int p = currentConf.nodeCount;
	
	/* determine & set the nodes own volume */
	if(p > nz)
	{
		thisNode.sites_on_nz = 1;
		
		if(p / nz > ny)
		{
			thisNode.sites_on_ny = 1;
			
			if(p / ny / nz > nx)
				thisNode.sites_on_nx = 1;
			else
				thisNode.sites_on_nx = nx * ny * nz / p;
		}
		else
			thisNode.sites_on_ny = ny * nz / p;
	}
	else
		thisNode.sites_on_nz = nz / p;
	
	/* determine the nodes offset position */
	v = thisNode.sites_on_nx * thisNode.sites_on_ny * thisNode.sites_on_nz;
	thisNode.sites_on_node = v;
	offV = thisNode.nodeIndex * v;
	V = 0;
	
	for(k = 0; k < nz / thisNode.sites_on_nz; k++)
	{
		for(j = 0; j < ny / thisNode.sites_on_ny; j++)
		{
			for(i = 0; i < nx / thisNode.sites_on_nx; i++)
			{
				if(V == offV)
				{
					thisNode.offset_x = i * thisNode.sites_on_nx;
					thisNode.offset_y = j * thisNode.sites_on_ny;
					thisNode.offset_z = k * thisNode.sites_on_nz;
					thisNode.parity = (thisNode.offset_x + thisNode.offset_y + thisNode.offset_z) % 2 == 1 ? ODD : EVEN;
                    debugPrint(ALL, "in determine Dimensions: nx=%i, ny=%i, nz=%i, xoff=%i, yoff=%i, zoff=%i", thisNode.sites_on_nx, thisNode.sites_on_ny, thisNode.sites_on_nz, thisNode.offset_x, thisNode.offset_y, thisNode.offset_z);
					return;
				}
				
				V += v;
			} /* i loop */
		} /* j loop */
	}  /* k loop */
}
