#include "../include/statistic.h"

/// <summary>
/// Initializes the statistic routines by allocating the required space
/// for building the current statistics
/// </summary>
/// <param name="reset">Determines if we just start a new round or if we start a new configuration</param>
void initializeStatistic(int reset)
{
	int n = currentConf.data.musteps * currentConf.data.betasteps;
	
	/* new configuration - set arrays */
	if(reset == ON)
	{
		/* free old memory */
		freeStatistic();
		
		/* create entries array when globalMaster */
		if(thisNode.globalIndex == 0)
		{
			debugPrint(GLOBALM, "Allocating space for statistic array with %d entries.", n);
			entries = (statisticEntry*)malloc(sizeof(statisticEntry) * n);
		}
	}
	
	/* n is now the number of configurations */
	debugPrint(LATTICEM, "Initializing the variables for the lattice statistics.");
	
	/* reseting the config stats */
	stat_S = 0.0;
	stat_Ssqu = 0.0;
	stat_H = 0.0;
	stat_Hsqu = 0.0;
	stat_N4 = 0.0;
}

/// <summary>
/// Frees the memory blocks that are allocated by the statistics
/// </summary>
void freeStatistic()
{
	/* if globalMaster and apparently some statistic set - free */
	if(thisNode.globalIndex == 0 && entries != NULL)
		free(entries);
}

/// <summary>
/// Calculates the current entries in this one specific lattice
/// configuration. This output is just temporary and will be 
/// needed to be transported to makeGlobalStatistic when all
/// repetitions are done.
/// </summary>
void updateLatticeStatistic()
{
    double sumSsqu = 0.0, sumS = 0.0, sumHsqu = 0.0, sumH = 0.0, sumN4 = 0.0, sumN4squ = 0.0;
	
	/* calculate statistics */
	double si = calculateSi();
	double n4i = calculateN4i();
	double hi = calculateHi(currentConf.mu);
	
	/* bring statistics together */
    vectorsReduceSum(&hi, &sumH, &sumHsqu);
    vectorsReduceSum(&si, &sumS, &sumSsqu);
    vectorsReduceSum(&n4i, &sumN4, &sumN4squ);
    
	/* sum up statistics */
    stat_S += sumS;
    stat_Ssqu += sumSsqu;
    stat_H += sumH;
    stat_Hsqu += sumHsqu;
	stat_N4 += sumN4;
    
    //debugPrint(LATTICEM, "Current S: %f, Ssq: %f, H: %f, Hsq: %f; N4: %f", stat_S, stat_Ssqu, stat_H, stat_Hsqu, stat_N4);
}

/// <summary>
/// Make the global statistic values i.e. saves the current entry in
/// the array that is later saved in the XML statistic evaluation file
/// This function goes over all lattices that means all configurations
/// i.e. all repetitions are brought together and averaged.
/// </summary>
/// <param name="r">The current entry that is produced</param>
void makeGlobalStatistic(int r)
{
	double global_Ssqu, global_S, global_H, global_Hsqu, global_N4; /* global variables */
	double sendmsg[5]; /* send vector */
	double recvmsg[5]; /* recv vector */
		
	/* create send / receive vector */
	sendmsg[0] = stat_S / currentConf.data.repetitions;
	sendmsg[1] = stat_Ssqu / currentConf.data.repetitions;
	sendmsg[2] = stat_H / currentConf.data.repetitions;
	sendmsg[3] = stat_Hsqu / currentConf.data.repetitions;
	sendmsg[4] = stat_N4 / currentConf.data.repetitions;
	
	debugPrint(LATTICEM, "Lattice S: %f,\tSsq: %f,\tH: %f,\tHsq: %f,\tN4: %f", sendmsg[0], sendmsg[1], sendmsg[2], sendmsg[3], sendmsg[4]);
	/* bring the lattices together */
	vectorsReduce(sendmsg, recvmsg, 5);

	/* only globalMaster does global Statistics */
	if(thisNode.globalIndex == 0)
	{            
		/* Assign the summed values */
		global_S = recvmsg[0];
		global_Ssqu = recvmsg[1];
		global_H = recvmsg[2];
		global_Hsqu = recvmsg[3];
		global_N4 = recvmsg[4];
		debugPrint(GLOBALM, "Global S: %f,\tSsq: %f,\tH: %f,\tHsq: %f,\tN4: %f", global_S, global_Ssqu, global_H, global_Hsqu, global_N4);
		
		/* storing for file save */
		entries[r].beta = currentConf.beta;
		entries[r].mu = currentConf.mu;
		entries[r].N3 =  currentConf.volume - global_N4;
		entries[r].N4 = global_N4;
		entries[r].H = global_H;
		entries[r].C = currentConf.beta * currentConf.beta * (global_Hsqu - (global_H * global_H));
		entries[r].X = currentConf.beta * currentConf.beta * (global_Ssqu - (global_S * global_S));
		
		/* also printing out directly if requested */
		if(currentConf.data.printstatistics == ON)
			masterPrint("Global stats for beta=%f, mu=%f: n3=%f, n4=%f, H=%f, C=%f, X=%f", 
				entries[r].beta, entries[r].mu, entries[r].N3, entries[r].N4, entries[r].H, entries[r].C, entries[r].X);
	}
}

/// <summary>
/// Saves the statistic output into a file named filename with a kind of XML notation
/// </summary>
/// <param name="fileName">The name of the file where the evaluation should be stored</param>
/// <returns>The error code of the file save operation</returns>
int saveStatistic(char *fileName)
{
	FILE *fp;
	char string[1024];
	time_t lt = time(NULL);
	struct tm *tmpstr = localtime(&lt);
	int length, i, n = currentConf.data.musteps * currentConf.data.betasteps;
	
	/* only globalMaster writes the file */
	if(thisNode.globalIndex == 0)
	{
		fp = fopen(fileName, "w+");
		
		/* write header */
		length = sprintf(string, "<?xml version=\"1.0\"?>\n");
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "<StatisticFile xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\">\n");
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<NX>%d</NX>\n", currentConf.data.nx);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<NY>%d</NY>\n", currentConf.data.ny);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<NZ>%d</NZ>\n", currentConf.data.nz);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<Entries>\n");
		fwrite(string, sizeof(char), length, fp);
		
		/* write values */
		for(i = 0; i < n; i++)
		{
			length = sprintf(string, "\t\t<StatisticEntry>\n");
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t\t<Beta>%f</Beta>\n", entries[i].beta);
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t\t<Mu>%f</Mu>\n", entries[i].mu);
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t\t<N4>%f</N4>\n", entries[i].N4);
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t\t<N3>%f</N3>\n", entries[i].N3);
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t\t<H>%f</H>\n", entries[i].H);
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t\t<C>%f</C>\n", entries[i].C);
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t\t<X>%f</X>\n", entries[i].X);
			fwrite(string, sizeof(char), length, fp);
			length = sprintf(string, "\t\t</StatisticEntry>\n");
			fwrite(string, sizeof(char), length, fp);
		}
		
		/* write footer */
		length = sprintf(string, "\t</Entries>\n");
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<Warmup>%d</Warmup>\n", currentConf.data.warmup);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<Iterations>%d</Iterations>\n", currentConf.data.iterations);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<Repetitions>%d</Repetitions>\n", currentConf.data.repetitions);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<DeltaBeta>%f</DeltaBeta>\n", currentConf.betastep);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<DeltaMu>%f</DeltaMu>\n", currentConf.mustep);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "\t<Time>%04d-%02d-%02dT%02d:%02d:%02d</Time>\n", tmpstr->tm_year + 1900, tmpstr->tm_mon + 1, 
						tmpstr->tm_mday, tmpstr->tm_hour, tmpstr->tm_min, tmpstr->tm_sec);
		fwrite(string, sizeof(char), length, fp);
		length = sprintf(string, "</StatisticFile>");
		fwrite(string, sizeof(char), length, fp);
		
		if(fclose(fp) != LATTICE_IO_NO_ERROR)
			return LATTICE_IO_NO_ACCESS;
	}
	
	return LATTICE_IO_NO_ERROR;
}

/// <summary>
/// Calculates the value for the statistic of sum_<x,y,z> s_i^2
/// </summary>
/// <returns>The current iterations value for sum s^2</returns>
double calculateN4i()
{
	int i, value = 0;
	
	/* loop over all sites and add s_{x,y,z}^2 */
	for(i = 0; i < thisNode.sites_on_node; i++)
		value += (thisNode.sites[i].value * thisNode.sites[i].value);
	
	return (double)value;
}

/// <summary>
/// Calculates the value for the statistic of sum_<x,y,z> s_i
/// </summary>
/// <returns>The current iterations value for sum s</returns>
double calculateSi()
{
	int i, value = 0;
	
	/* loop over all sites and add s_{x,y,z}^2 */
	for(i = 0; i < thisNode.sites_on_node; i++)
		value += thisNode.sites[i].value;

	return (double)value;
}

/// <summary>
/// Calculates the value for the statistic of sum_<x,y,z> h_i where h
/// is the (little) hamiltonian i.e. the one that only lives on this sublattice
/// </summary>
/// <param name="mu">The current value of the parameter mu</param>
/// <returns>The current iterations value for sum h</returns>
double calculateHi(double mu)
{
	int i;
	double value = 0.0;
	
	/* loop over all sites and add h_{x,y,z} = s_{x,y,z} * sum_neighbors + mu * s_{x,y,z}^2 */
	for(i = 0; i < thisNode.sites_on_node; i++)
		value -= ((double)(thisNode.sites[i].value * getNeighborSum(i)) + mu * (double)(thisNode.sites[i].value * thisNode.sites[i].value));
    
    return value;
}

