#include "../include/lattice_io.h"

/// <summary>
/// Saves the lattice in text or binary form (text not supported yet)
/// </summary>
/// <param name="fileName">The name of the file where the lattice should be stored</param>
/// <param name="txtOrBinary">A flag specifying in which mode to save the file</param>
/// <returns>Returns the error code of the file save operation</returns>
int saveLattice(char *fileName, int txtOrBinary)
{
	int k, x, y, z;
	time_t lt = time(NULL);
	struct tm *tmpstr = localtime(&lt);
	FILE *fp = NULL;
#ifdef MPICOMM
	int offset = thisNode.nodeIndex * (thisNode.sites_on_nx + thisNode.sites_on_ny + thisNode.sites_on_nz) * sizeof(char);
	MPI_File ufh;
	MPI_Status status;
#endif

	/* writes the header */
	if(thisNode.globalIndex == 0)
	{
		fp = fopen(fileName, "w+");

		/* Textmode at the moment not supported */
		if(txtOrBinary == LATTICE_IO_BINARYMODE || txtOrBinary == LATTICE_IO_TEXTMODE)
		{
			int yr = tmpstr->tm_year + 1900;
			int mo = tmpstr->tm_mon + 1;

			/* write header */
			fwrite(&(tmpstr->tm_mday), sizeof(int), 1, fp);
			fwrite(&mo, sizeof(int), 1, fp);
			fwrite(&yr, sizeof(int), 1, fp);
			fwrite(&(tmpstr->tm_hour), sizeof(int), 1, fp);
			fwrite(&(tmpstr->tm_min), sizeof(int), 1, fp);
			fwrite(&(tmpstr->tm_sec), sizeof(int), 1, fp);
			fwrite(&(currentConf.data.nx), sizeof(int), 1, fp);
			fwrite(&(currentConf.data.ny), sizeof(int), 1, fp);
			fwrite(&(currentConf.data.nz), sizeof(int), 1, fp);
			fwrite(&(currentConf.beta), sizeof(double), 1, fp);
			fwrite(&(currentConf.mu), sizeof(double), 1, fp);
		}
		
#ifdef MPICOMM
		/* MPI call only - after that we work with the MPI_File - therefore close FILE */
		if(fclose(fp) != LATTICE_IO_NO_ERROR)
			return LATTICE_IO_NO_ACCESS;
#endif
	}
	
	COMM_Barrier();

	debugPrint(GLOBALM, "Starting writing the file...");

	/* writes the lattice - since only one can be saved lets save config 0 (1st)*/
	if(currentConf.latticeIndex == 0)
	{
#ifdef MPICOMM
		for(offset = 0; offset < currentConf.nodeCount; offset++)
		 {
			if(thisNode.nodeIndex == offset)
			 {
		/* since we work now with MPI_File we have to open and set view */
				MPI_File_open(MPI_COMM_SELF, fileName, MPI_MODE_APPEND | MPI_MODE_RDWR, MPI_INFO_NULL, &ufh);
				//MPI_File_set_view (ufh, offset, MPI_CHAR, MPI_CHAR, "native", MPI_INFO_NULL);
#endif
		k = 0;
		for(z = 0; z < thisNode.sites_on_nz; z++)
		{
			for(y = 0; y < thisNode.sites_on_ny; y++)
			{
				for(x = 0; x < thisNode.sites_on_nx; x++)
				{
					/* Textmode at the moment not supported */
					if(txtOrBinary == LATTICE_IO_BINARYMODE || txtOrBinary == LATTICE_IO_TEXTMODE)
					{
                        			char c = (char)(thisNode.sites[k].value);
#ifdef MPICOMM
						MPI_File_write(ufh, &c, 1, MPI_CHAR, &status);
#else
						
						fwrite(&c, sizeof(char), 1, fp);
#endif /* MPICOMM */
					}
					
					k++;
				} /* x */
			} /* y */
		} /* z */
		
#ifdef MPICOMM
			}
			COMM_Barrier();
		}
		/* just in MPI we have more than one parallel configuration running */
		MPI_File_close(&ufh);
#endif
	}

	debugPrint(GLOBALM, "Finished writing the file...");

	COMM_Barrier();

#ifndef MPICOMM
	/* FILE is still open if we are not in MPICOMM mode */
	if(fclose(fp) != LATTICE_IO_NO_ERROR)
		return LATTICE_IO_NO_ACCESS;
#endif
	
	return LATTICE_IO_NO_ERROR;		
}

/// <summary>
/// Loads the lattice from a text or binary file (text not supported yet)
/// </summary>
/// <param name="fileName">The name of the file where the lattice should be loaded from</param>
/// <returns>Returns the error code of the file load operation</returns>
int loadLattice(char *fileName)
{
	int err, k = 0, x, y, z, v = thisNode.sites_on_nx * thisNode.sites_on_ny * thisNode.sites_on_nz;
#ifdef MPICOMM
	int offset = thisNode.nodeIndex * v + 9 * sizeof(int) + 2 * sizeof(double);
	MPI_File ufh;
	MPI_Status status;
#else
	FILE *fp;
#endif /* MPICOMM */
	
	if(thisNode.globalIndex == 0)
	{
		err = readLatticeHeader(fileName, ON);
		
		if(err != LATTICE_IO_NO_ERROR)
		{
			if(err == LATTICE_IO_WRONG_DIM)
				masterPrint("The dimensions of the loaded lattice and the configured one do not match.");
				
			terminate(EXIT_FILE_ERR);
		}
	}
	
	COMM_Barrier();
#ifdef MPICOMM
	MPI_File_open(MPI_COMM_WORLD, fileName, MPI_MODE_RDONLY, MPI_INFO_NULL, &ufh);
	MPI_File_set_view (ufh, offset, MPI_CHAR, MPI_CHAR, "native", MPI_INFO_NULL);
#else
	fp = fopen(fileName, "r");
	
	if(fp == NULL)
		return LATTICE_IO_INV_FILENAME;
#endif /* MPICOMM */

	/* set the site count */
	thisNode.sites = (site*)malloc(sizeof(site) * v);
	
	/* loop over all the node's sites */
	for(z = 0; z < thisNode.sites_on_nz; z++)
	{
		for(y = 0; y < thisNode.sites_on_ny; y++)
			for(x = 0; x < thisNode.sites_on_nx; x++)
			{
				char value;
				thisNode.sites[k].x = x;
				thisNode.sites[k].y = y;
				thisNode.sites[k].z = z;
				thisNode.sites[k].parity = (x + y + z) % 2 == 1 ? ODD : EVEN;
				
				/* loads the value */
#ifdef MPICOMM
				MPI_File_read(ufh, &value, 1, MPI_CHAR, &status);
#else
				fread(&value, sizeof(char), 1, fp);
#endif /* MPICOMM */
				thisNode.sites[k].value = (int)value;
				k++;
			} /* x */
	} /* z loop */
	
#ifdef MPICOMM
	MPI_File_close(&ufh);
#else
	if(fclose(fp) != LATTICE_IO_NO_ERROR)
		return LATTICE_IO_NO_ACCESS;
#endif /* MPICOMM */

	COMM_Barrier();
	return LATTICE_IO_NO_ERROR;
}

/// <summary>
/// Reads the header information of a lattice file that is about to be loaded
/// </summary>
/// <param name="fileName">The name of the file that contains the header information</param>
/// <param name="checkSize">A flag that states of the volume size should be checked (this flag is on for a whole lattice load and off for a just load the header information procedure)</param>
/// <returns>Returns the error code of the file load operation</returns>
int readLatticeHeader(char* fileName, int checkSize)
{
	int nx, ny, nz, length; /* lattice and file size */
	double beta, mu; /* parameters */
	int yr, mo, dy, hr, mi, se; /* datetime variables */
	int err = LATTICE_IO_NO_ERROR; /* we do not expect an error */
	int offset = 2 * sizeof(double) + 9 * sizeof(int);
	FILE *fp = fopen(fileName, "r");
	
	/* if file is open OK */
	if(fp != NULL)
	{
		fseek(fp, 0L, SEEK_END);
		length = ftell(fp);
		
		/* if length is OK i.e. header possible */
		if(length > offset)
		{
			/* reset position */
			fseek(fp, 0L, SEEK_SET);
			/* read header */
			fread(&dy, sizeof(int), 1, fp);
			fread(&mo, sizeof(int), 1, fp);
			fread(&yr, sizeof(int), 1, fp);
			fread(&hr, sizeof(int), 1, fp);
			fread(&mi, sizeof(int), 1, fp);
			fread(&se, sizeof(int), 1, fp);
			fread(&nx, sizeof(int), 1, fp);
			fread(&ny, sizeof(int), 1, fp);
			fread(&nz, sizeof(int), 1, fp);
			fread(&beta, sizeof(double), 1, fp);
			fread(&mu, sizeof(double), 1, fp);
			
			/* print header */
			printf("The following information is from the provided lattice file:\n");
			printf("nx:\t%d\n", nx);
			printf("ny:\t%d\n", ny);
			printf("nz:\t%d\n", nz);
			printf("beta:\t%f\n", beta);
			printf("mu:\t%f\n", mu);
			printf("INFO: The lattice has been saved on %02d.%02d.%04d at %02d:%02d,%02d.\n", dy, mo, yr, hr, mi, se);
			
			/* compare file size */
			if(nx * ny * nz * sizeof(char) + offset != length)
			{
				printf("ERROR: The file size does not match the lattice volume. The file seems corrupt.\n");
				err = LATTICE_IO_WRONG_DIM;
			}
			
			if(checkSize == ON)
				err = (nx == currentConf.data.nx && ny == currentConf.data.ny && nz == currentConf.data.nz) ? LATTICE_IO_NO_ERROR : LATTICE_IO_WRONG_DIM;
		}
		else
		{
			char msg[1024] = "The provided lattice file contains an invalid header.\n";
			masterPrint(msg);
			err = LATTICE_IO_ERR_HEAD;
		}
	}
	else
	{
		char msg[1024] = "The provided lattice file is not accessible (does the file exist?)\n";
		masterPrint(msg);
		return LATTICE_IO_NO_ACCESS;
	}
	
	/* close the file */
	fclose(fp);
	return err;
}

