#include "../include/comm.h"

// Init and finalize
/// <summary>
/// The main function of our Tricritical He3/He4 Mixing Ising model for the HPSC lecture
/// </summary>
/// <param name="argc">number of command line arguments</param>
/// <param name="argv">command line arguments as strings</param>
void initialize(int argc, char* argv[])
{
#ifdef MPICOMM
    debugPrint(ALL,"MPICOMM is defined");
	int blade_cpus = BLADE_CPUS;
#endif
	thisNode.nodeIndex = 0;
	currentConf.latticeIndex = 0;
	latticeCount = 1;
#ifdef MPICOMM
	// execute MPI code
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &(thisNode.globalIndex));
	MPI_Comm_size(MPI_COMM_WORLD, &processCount);
	
#ifndef BLADE_SPLIT
	blade_cpus = processCount;
#endif /*BLADE_SPLIT*/
	thisNode.nodeIndex = thisNode.globalIndex % blade_cpus;
    
	if(processCount > blade_cpus)
	{
		if(processCount % blade_cpus != 0)
			terminate(EXIT_CPU_COUNT);
        
		currentConf.nodeCount = blade_cpus;
        latticeCount = processCount / blade_cpus;
        currentConf.latticeIndex = thisNode.globalIndex / blade_cpus;
	}
	else
		currentConf.nodeCount = processCount;
    
    debugPrint(ALL,"in initialize: globalIndex: %i, nodeIndex: %i, blade_cpus: %i, latticeIndex:%i", thisNode.globalIndex, thisNode.nodeIndex, blade_cpus, currentConf.latticeIndex);
    
    
    //
    // Make sub topologies.
    //
    /*MPI_Group groupWorld;
    int* bladeProcesses;
    int* masterProcesses;
    // Make a list of processes in the new communicator
    bladeProcesses = (int*) malloc(currentConf.blade_cpus*sizeof(int));
    masterProcesses = (int*) malloc(latticeCount*sizeof(int));
    int proc;
    for (proc = 0; proc < currentConf.blade_cpus; proc++)
    {
        bladeProcesses[proc] = proc;
    }
    for (proc = 0; proc < latticeCount; proc++)
    {
        bladeProcesses[proc] = proc;
    }
    //Get the group underlying MPI_COMM_WORLD
    MPI_Comm_group(MPI_COMM_WORLD, &groupWorld);
    //Create the new group
    MPI_Group_incl(groupWorld, currentConf.blade_cpus, bladeProcesses, &currentConf.MPI_COMM_BLADE);
    MPI_Group_incl(groupWorld, latticeCount, masterProcesses, &currentConf.MPI_COMM_MASTERS);
    // Create the new communicator
    MPI_Comm_create(MPI_COMM_WORLD, currentConf.MPI_COMM_BLADE, &MPI_COMM_BLADE);*/
    
    //
    // Make sub topologies (faster).
    //
    // makes Blade Communicator. Put all with same latticeIndex in one Comm
    MPI_Comm_split(MPI_COMM_WORLD, currentConf.latticeIndex, thisNode.globalIndex, &thisNode.MPI_COMM_BLADE);
    // MPI_Comm_split(MPI_COMM_WORLD, thisNode.nodeIndex, thisNode.globalIndex, &thisNode.MPI_COMM_BLADE);
    // makes Master Communicator. Put all with same nodeIndex in one Comm
    MPI_Comm_split(MPI_COMM_WORLD, thisNode.nodeIndex, thisNode.globalIndex, &thisNode.MPI_COMM_MASTERS);
    
    
    
#else
    debugPrint(ALL, "MPICOMM is not defined");
	processCount = 1;
	thisNode.globalIndex = 0;
	currentConf.nodeCount = 1;
#endif /*MPICOMM*/
}

/// <summary>
/// Terminates the program with the termination code
/// <summary>
/// <param name="status">The termination code / program exit code that will be send</param>
void terminate(int status)
{
	freeStatistic();
#ifdef MPICOMM
	// execute MPI code
	MPI_Finalize();
#endif /*MPICOMM*/

	if(status != EXIT_OK)
		printf("Termination: node %d, status = %d\n", thisNode.nodeIndex, status);
	
	fflush(stdout);
	exit(status);
}

/// <summary>
/// Performs the sending of a message over MPI with doubles
/// <summary>
/// <param name="message">The array with double data to be send</param>
/// <param name="count">The length of the array to be send</param>
/// <param name="dest">The target processor of the current message</param>
/// <param name="tag">The tag of the message to identify it</param>
/// <returns>The MPI_Send error code</returns>
int COMM_Send(void* message, int count, int dest, int tag)
{
    #ifdef MPICOMM
        // execute MPI code
        return MPI_Send(message, count, MPI_DOUBLE, dest, tag, MPI_COMM_WORLD);
    #else
        //no communication needed.
        return 0;
    #endif /*MPICOMM*/
}

/// <summary>
/// Performs the receiving of a message over MPI with doubles
/// <summary>
/// <param name="message">The array with double data to be received</param>
/// <param name="count">The length of the array to be received</param>
/// <param name="dest">The source processor of the current message</param>
/// <param name="tag">The tag of the message to identify it</param>
/// <returns>The MPI_Recv error code</returns>
int COMM_Recv(void* message, int count, int source, int tag)
{
#ifdef MPICOMM
	// execute MPI code
	MPI_Status  status;        // return status for receive
	return MPI_Recv(message, count, MPI_DOUBLE, source, tag, MPI_COMM_WORLD, &status);
#else
	//execute code w/o makefile or generally
	return 0;
#endif /*MPICOMM*/	
}

/// <summary>
/// Performs the collective operation of a broadcast
/// <summary>
/// <param name="message">The array with double data to be send / recv</param>
/// <param name="count">The length of the array to be send /recv</param>
/// <param name="root">The sender of the broadcast (usually a lattice master or even the global master)</param>
/// <returns>The MPI_Bcast error code</returns>
int COMM_Bcast(void* message, int count, int root)
{
#ifdef MPICOMM
	// execute MPI code
	return MPI_Bcast(message, count, MPI_DOUBLE, root, MPI_COMM_WORLD);
#else
	//execute code w/o makefile or generally
	return 0;
#endif /*MPICOMM*/
}

/// <summary>
/// Performs the collective operation of a gather
/// <summary>
/// <param name="sendbuf">The data that will be send with the gather</param>
/// <param name="sendcount">The length of the array to be send</param>
/// <param name="recvbuf">The data that will be received with the gather</param>
/// <param name="recvcount">The length of the array to be received</param>
/// <param name="root">The root who does the gather job - usually lattice master</param>
/// <returns>The MPI_Gather error code</returns>
int COMM_Gather(void* sendbuf, int sendcount, void* recvbuf, int recvcount, int root)
{
#ifdef MPICOMM
	// execute MPI code
	return MPI_Gather(sendbuf, sendcount, MPI_DOUBLE, recvbuf, recvcount, MPI_DOUBLE, root, MPI_COMM_WORLD);
#else
	//execute code w/o makefile or generally
	return 0;
#endif /*MPICOMM*/
}

/// <summary>
/// Performs the collective operation of an allgather
/// <summary>
/// <param name="sendbuf">The data that will be send with the gather</param>
/// <param name="sendcount">The length of the array to be send</param>
/// <param name="recvbuf">The data that will be received with the gather</param>
/// <param name="recvcount">The length of the array to be received</param>
/// <returns>The MPI_Allgather error code</returns>
int COMM_Allgather(void* sendbuf, int sendcount, void* recvbuf, int recvcount)
{
#ifdef MPICOMM
	// execute MPI code
	return MPI_Allgather(sendbuf, sendcount, MPI_DOUBLE, recvbuf, recvcount, MPI_DOUBLE, MPI_COMM_WORLD);
#else
	//execute code w/o makefile or generally
	return 0;
#endif /*MPICOMM*/
}

/// <summary>
/// Performs the collective operation of a reduce (with summing up the vector)
/// <summary>
/// <param name="operand">The data that will be send</param>
/// <param name="result">The result in form of a double</param>
/// <param name="count">The length of the output</param>
/// <param name="root">The processor that performs the job - usually lattice master</param>
/// <returns>The MPI_Reduce error code</returns>
int COMM_ReduceSUM(void* operand, void* result, int count, int root)
{
#ifdef MPICOMM
	// execute MPI code
	return MPI_Reduce(operand, result, count, MPI_DOUBLE, MPI_SUM, root, MPI_COMM_WORLD);
#else
	//execute code w/o makefile or generally
	return 0;
#endif /*MPICOMM*/
}

/// <summary>
/// Performs the collective operation of an allreduce (with summing up the vector)
/// <summary>
/// <param name="operand">The data that will be send</param>
/// <param name="result">The result in form of a double</param>
/// <param name="count">The length of the output</param>
/// <returns>The MPI_Allreduce error code</returns>
int COMM_AllreduceSUM(void* operand, void* result, int count)
{
#ifdef MPICOMM
	// execute MPI code
	return MPI_Allreduce(operand, result, count, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
#else
	//execute code w/o makefile or generally
	return 0;
#endif /*MPICOMM*/
}

/// <summary>
/// Distributes the inputData structure (currentConf.data) over all the nodes - every
/// must have the same configuration in order to produce reliable results
/// <summary>
void distributeInputData()
{
    #ifdef MPICOMM
        // Broadcast input data. Process 0 sends, all other processes receive the data.
        MPI_Bcast(&currentConf.data, sizeof(struct inputData), MPI_CHAR, 0, MPI_COMM_WORLD);
    #else
        // nothing.
    #endif /*MPICOMM*/
}

/// <summary>
/// Performs a barrier i.e. everyone waits until everyone is finished
/// <summary>
void COMM_Barrier()
{
#ifdef MPICOMM
	MPI_Barrier(MPI_COMM_WORLD);
#endif /*MPICOMM*/
}
