#include "../include/main.h"

/// <summary>
/// The main function of our Tricritical He3/He4 Mixing Ising model for the HPSC lecture
/// all copyright (c) Christoph Preis and Florian Rappl
/// Boulder, Colorado in April 2011.
/// </summary>
/// <param name="argc">number of command line arguments</param>
/// <param name="argv">command line arguments as strings</param>
/// <returns>integer that specifies termination code (0 = no error)</returns>
int main(int argc, char* argv[])
{
#ifndef MPICOMM
	/* So we can measure time in single mode if we want to */
	masterPrint("Start.");
#endif
    
    /* General Variables */
    int i, r, b, m, w, reset; /* loop counters */
    int prompt, oldprompt; /* if input prompt should be shown */
#ifdef MPICOMM
	double starttime, endtime; /* track time in MPI mode */
#endif
    
	/* Start up communication; will or will not run MPI depending on flag MPICOMM */
	initialize(argc, argv);
    
    debugPrint(GLOBALM, "HPSC FINAL PROJECT");
    debugPrint(GLOBALM, "Communication initialized!");
	
	/* parse the commandLine (read given input file ?) or shown input prompt  (no input file given) */
    debugPrint(GLOBALM, "Initializing command line parser...");
	prompt = commandLineParser(argc, argv);
	
	do
	{
#ifdef MPICOMM
		starttime = MPI_Wtime();
#endif
		/* sets the continue mode to the prompt value */
		oldprompt = prompt;
		
		/* set current lattice variables according to configuration variables */
		reset = ON;
		currentConf.volume = currentConf.data.nx * currentConf.data.ny * currentConf.data.nz;
		currentConf.mu = currentConf.data.mustart;
		currentConf.mustep = (currentConf.data.muend - currentConf.data.mustart) / (double)currentConf.data.musteps;
		currentConf.betastep = (currentConf.data.betaend - currentConf.data.betastart) / (double)currentConf.data.betasteps;
		
		/* create the node's sub lattice or read it */
        debugPrint(GLOBALM, "Setting sublattice...");
		setSubLattice();
		
		/* tells the update routine to start - sets all indices to 0 and exchanges ghost sites */
        debugPrint(GLOBALM, "Initializing update routine...");
		initializeUpdate();
		
		/* main loop over given configurations */
        debugPrint(GLOBALM, "MAIN LOOP...");
		for (m = 0; m < currentConf.data.musteps; m++)
		{
            debugPrint(GLOBALM, "MAIN LOOP: New mu.");
			currentConf.beta = currentConf.data.betastart;

			for (b = 0; b < currentConf.data.betasteps; b++)
			{
                debugPrint(GLOBALM, "MAIN LOOP: New beta.");
				/* create a lookup table so that we have need less computation */
				double** lookup = createLookup(currentConf.beta, currentConf.mu);
				
				/* initialize all vectors and structs that are required to measure */
				initializeStatistic(reset);
				reset = OFF;
				
				/* perform the specified number of warmup runs */
				debugPrint(GLOBALM, "Starting warmups...");
				for(w = 0; w < currentConf.data.warmup; w++)
					update(lookup, currentConf.beta, currentConf.mu);
				
				/* perform the specified number of iterations */
				debugPrint(GLOBALM, "Starting iterations...");
				for(r = currentConf.latticeIndex; r < currentConf.data.repetitions; r += latticeCount)
				{
					for(i = 0; i < currentConf.data.iterations; i++)
						update(lookup, currentConf.beta, currentConf.mu);
					
					/* bring together - measure global */
					updateLatticeStatistic();				
				}
				
				/* average over all configurations - smoothen / make more accurate */
                debugPrint(GLOBALM, "Updating global stats.");
				makeGlobalStatistic(m * currentConf.data.betasteps + b);
				currentConf.beta += currentConf.betastep;
			}

			currentConf.mu += currentConf.mustep;
		}
		
#ifdef MPICOMM
		/* calculating calculation time ... */
		endtime = MPI_Wtime();
		masterPrint("The time for this computation was: %lf", endtime - starttime);
#endif
		
		/* save the statistic (evaluation) file */
        debugPrint(GLOBALM, "Saving statistics file.");
		if(currentConf.data.saveeval == YES_SAVE_EVAL)
			if(saveStatistic(currentConf.data.saveevalfile) != LATTICE_IO_NO_ERROR)
			{
				masterPrint("The statistic file could not been written. Check your configuration.");
				terminate(EXIT_FILE_ERR);
			}
		
		/* save the current lattice (if so configured) */
		switch(currentConf.data.savelattice)
		{
			case SAVE_LATTICE_DEFAULT:
			case SAVE_LATTICE_BINARY:
				if(currentConf.data.loadlattice != LOAD_LATTICE_TEXT)
				{
					/* Call save binary */
                    debugPrint(GLOBALM, "Saving lattice file as binary.");
					saveLattice(currentConf.data.savelatticefile, LATTICE_IO_BINARYMODE);
					break;
				}
			case SAVE_LATTICE_TEXT:
				/* Call save text */
                debugPrint(GLOBALM, "Saving lattice file as text.");
				saveLattice(currentConf.data.savelatticefile, LATTICE_IO_TEXTMODE);
				break;
			case FORGET_LATTICE:
				/* Forgets about the lattice file */
                debugPrint(GLOBALM, "Discarding lattice file.");
				break;
		}
		
		/* show input prompt for further evaluations */
		if(oldprompt == SHOW_PROMPT)
			prompt = inputPrompt();
			
	} while(oldprompt == SHOW_PROMPT);
	
#ifndef MPICOMM
	/* So we can measure time in single mode if we want to */
	masterPrint("Exit.");
#else
    debugPrint(GLOBALM, "Exit.");	
#endif

	/* exits the program */
	terminate(EXIT_OK);	
	return 0;
}

