#include "../include/update.h"

/* global helpers - saves computation power */
int *volumes;		/* small volumes (areas) */
int **sendRecvSites;/* values of the sites to send and recv */
int *p;				/* number of processors in each direction */
int *nofs;			/* offset when communicating with nearest p processor */

/// <summary>
/// Initializes the update routines i.e. allocates space for variables
/// and sets starting variables
/// <summary>
void initializeUpdate()
{
    debugPrint(ALL, "Entering update routine. My nodeIndex is %i.", thisNode.nodeIndex);
	int i, j;
	/* set starting variables */
	updateIndex = 0;
	updateParity = EVEN;
	
	if(currentConf.data.iseed == 0)
		srand((unsigned)time(NULL));
	else
		srand((unsigned)currentConf.data.iseed);
	
	/* free old ghostSites */
	if(thisNode.ghostSites != NULL)
	{
		for(i = 0; i < 6; i++)
			free(thisNode.ghostSites[i]);
			
		free(thisNode.ghostSites);
	}
		
	thisNode.ghostSites = (site**)malloc(sizeof(site*) * 6);
	volumes = (int*)malloc(sizeof(int) * 6);
	sendRecvSites = (int**)malloc(sizeof(int*) * 6); 
	p = (int*)malloc(sizeof(int) * 6); 
	nofs = (int*)malloc(sizeof(int) * 6);
	
	/* fill ghostsite array */
	for(i = LEFT; i <= DOWN; i++)
	{
		/* determines the small volume */
		int x = i == LEFT || i == RIGHT ? 1 : thisNode.sites_on_nx;
		int y = i == TOP || i == BOTTOM ? 1 : thisNode.sites_on_ny;
		int z = i == UP || i == DOWN ? 1 : thisNode.sites_on_nz;
		
		volumes[i] = x * y * z;
		thisNode.ghostSites[i] = (site*)malloc(sizeof(site) * volumes[i]);
		sendRecvSites[i] = (int*)malloc(sizeof(int) * volumes[i]);
		
		/* some starting values - just in case and for ni = 1 */
		for(j = 0; j < volumes[i]; j++)
			sendRecvSites[i][j] = 0;
			
		/* sets the processor counts and neighbor offsets */
		if(i == LEFT)
		{
			p[i] = currentConf.data.nx / thisNode.sites_on_nx;
			nofs[i] = 1;
		}
		else if(i == TOP)
		{
			p[i] = currentConf.data.ny / thisNode.sites_on_ny;
			nofs[i] = p[LEFT];
		}
		else if(i == UP)
		{
			p[i] = currentConf.data.nz / thisNode.sites_on_nz;
			nofs[i] = p[LEFT] * p[TOP];
		}
		else
		{
			p[i] = p[i - 3];
			nofs[i] = nofs[i - 3];
		}
	}
		
	/* exchange them for first time (set real values) */
	exchangeGhostsites();
}

/// <summary>
/// Exchanges all the site values that represent the boundary
/// values of the node's sublattice
/// <summary>
void exchangeGhostsites()
{
#ifdef MPICOMM
	MPI_Status status;
	int partner; /* partner to send / recv */
#endif
	int *tmpvec;
	int i, j, l = 0, t = 0, u = 0, r = 0, b = 0, d = 0, k, f;
	int mirror; /* mirror to store */
	
	debugPrint(LATTICEM, "EXCHANGING GHOSTSITES");
	
	/* fill the send sites */
	for(j = 0; j < thisNode.sites_on_node; j++)
	{
		/* fill something with x */
		if(thisNode.sites[j].x == 0 || thisNode.sites[j].x == thisNode.sites_on_nx - 1)
		{ 
			/* we just have x = 1 --> left = right unless its no real dim! then we set 0 */
			if(thisNode.sites_on_nx == 1)
			{
				k = currentConf.data.nx == 1 ? 0 : thisNode.sites[j].value;
				sendRecvSites[LEFT][l++] = k;
				sendRecvSites[RIGHT][r++] = k;
			} /* we just have x > 1 --> determine if it was left or right */
			else if(thisNode.sites[j].x == 0)
				sendRecvSites[LEFT][l++] = thisNode.sites[j].value;
			else
				sendRecvSites[RIGHT][r++] = thisNode.sites[j].value;			
		} /* fill something with y */
		else if(thisNode.sites[j].y == 0 || thisNode.sites[j].y == thisNode.sites_on_ny - 1)
		{
			/* we just have y = 1 --> top = bottom unless its no real dim! then we set 0 */
			if(thisNode.sites_on_ny == 1)
			{
				k = currentConf.data.ny == 1 ? 0 : thisNode.sites[j].value;
				sendRecvSites[TOP][t++] = k;
				sendRecvSites[BOTTOM][b++] = k;
			} /* we just have y > 1 --> determine if it was top or bottom */
			else if(thisNode.sites[j].y == 0)
				sendRecvSites[TOP][t++] = thisNode.sites[j].value;
			else
				sendRecvSites[BOTTOM][b++] = thisNode.sites[j].value;
		} /* fill something with z */
		else if(thisNode.sites[j].z == 0 || thisNode.sites[j].z == thisNode.sites_on_nz - 1)
		{
			/* we just have z = 1 --> up = down unless its no real dim! then we set 0 */
			if(thisNode.sites_on_nz == 1)
			{
				k = currentConf.data.nz == 1 ? 0 : thisNode.sites[j].value;
				sendRecvSites[UP][u++] = k;
				sendRecvSites[DOWN][d++] = k;
			} /* we just have z > 1 --> determine if it was up or down */
			else if(thisNode.sites[j].z == 0)
				sendRecvSites[UP][u++] = thisNode.sites[j].value;
			else
				sendRecvSites[DOWN][d++] = thisNode.sites[j].value;
		}	
	}
	
	COMM_Barrier();
	
	/* do the exchange one time all even other time all odd are sending*/
	for(i = LEFT; i <= 2 * DOWN + 1; i++)
	{
        COMM_Barrier();
		j = i / 2;
		f = j < 3 ? 1 : -1;
        debugPrint(ALL, "in exchange Ghostsites: i=%i, j=%i, f=%i", i, j, f );
		
		/* determines if we processor is even or odd in the current direction */
		if((thisNode.nodeIndex / nofs[j] + thisNode.nodeIndex % nofs[j]) % 2 == i % 2)
		{
#ifdef MPICOMM
			if(p[j] != 1)
			{
                debugPrint(ALL, "in exchange Ghostsites: calculation of partner: lattice index=%i, nodeCount=%i, nodeIndex=%i, f=%i, nofs=%i, p=%i, dir=%i", currentConf.latticeIndex, currentConf.nodeCount, thisNode.nodeIndex, f, nofs[j], p[j], j);
				partner = (thisNode.nodeIndex - f * nofs[j] + p[j]) % p[j];
                debugPrint(ALL, "in exchange Ghostsites: nodeIndex %i sends to %i", thisNode.nodeIndex, partner);
				MPI_Send(sendRecvSites[j], volumes[j], MPI_INT, partner, j, thisNode.MPI_COMM_BLADE);
			} /*send --> sendRecvSites[j] */
#endif
		}
		else
		{
			tmpvec = (int*)malloc(sizeof(int) * volumes[j]);
			
#ifdef MPICOMM
			if(p[j] != 1)
			{
				partner = (thisNode.nodeIndex + f * nofs[j] + p[j]) % p[j];
                debugPrint(ALL, "in exchange Ghostsites: nodeIndex %i recv from %i, dir=%i", thisNode.nodeIndex, partner, j);
                MPI_Recv(tmpvec, volumes[j], MPI_INT, partner, j, thisNode.MPI_COMM_BLADE, &status);
			} /*recv --> sendRecvSites[j] */
#else
			for(k = 0; k < volumes[j]; k++)
				tmpvec[k] = sendRecvSites[j][k];
#endif
			
			/* recv top -> make bottom ghostrow etc. */
			mirror = (j + 3) % 6; /* we need a mirror */
			
			/* fill ghostsite array */
			for(k = 0; k < volumes[j]; k++)
				thisNode.ghostSites[mirror][k].value = tmpvec[k];
				
			free(tmpvec);
		}
        COMM_Barrier();
	}
	
	debugPrint(LATTICEM, "EXCHANGE DONE");
}

/// <summary>
/// Runs an update process on the current update index
/// <summary>
/// <param name="lookup">the table with all the possible probability values</param>
/// <param name="beta">current iteration parameter beta</param>
/// <param name="mu">current iteration parameter mu</param>
void update(double** lookup, double beta, double mu)
{
	int delta = getNeighborSum(updateIndex), i;
	double total = 0.0;
	double prob = (double)(rand() % 101) / 100.0;
	
	//printf("probability = %f\tneighbors = %d\n", prob, delta);

	for (i = -1; i < 2; i++)
	{
		total += lookup[i + 1][delta + 6];
		
		if (prob <= total)
		{
			//printf("IDX %d accepted change (from, to)\t (%d, %d)\n", updateIndex, thisNode.sites[updateIndex].value, i);
			thisNode.sites[updateIndex].value = i;
			break;
		}
	}
	
	updateIndex += 2;
	
	if(updateIndex >= thisNode.sites_on_node)
	{
		updateParity = updateParity == EVEN ? ODD : EVEN;
		updateIndex = updateParity % 2;
		exchangeGhostsites();
	}
}

/// <summary>
/// Creates the table with all possible probability values for the specific beta,mu pair
/// First dimension is the current sites value, second dimension all 13 neighbor possibilities
/// with the current sites value (-1 0 1) from -6 .. 0 .. +6
/// <summary>
/// <param name="beta">current iteration parameter beta</param>
/// <param name="mu">current iteration parameter mu</param>
/// <returns>The two dimensional lookup table</returns>
double** createLookup(double beta, double mu)
{
	int i, j;
	double norm;
	double **lookup = (double**)malloc(sizeof(double*) * 3);
	
	/* allocate space for lookup table 3x13 */
	/* since [-1 0 +1] on the site and [-6 ... +6] from neighbors */
	for(i = 0; i < 3; i++)
		lookup[i] = (double*)malloc(sizeof(double) * 13);

	for (i = -6; i < 7; i++)
	{
		/* reset the norm */
		norm = 0.0;
		
		/* compute and sum for normalization factor */
		for (j = -1; j < 2; j++)
		{
			lookup[j + 1][i + 6] = exp(beta * (double)j * ((double)i + mu * (double)j));
			norm += lookup[j + 1][i + 6];
		}
		
		/* normalize */
		for (j = -1; j < 2; j++)
			lookup[j + 1][i + 6] = lookup[j + 1][i + 6] / norm;
	}
	
	/* code to debug lookup table */
	/*
	debugPrint(GLOBALM, "The lookup table for b=%f m=%f", beta, mu);
	for(j = 0; j < 3; j++)
	{
		printf("%d:\t", j - 1);
		
		for(i = 0; i < 13; i++)
			printf("%f\t", lookup[j][i]);
			
		printf("\n");
	}
	*/
	
	return lookup;
}

/// <summary>
/// Gets the sum of the neighbor sites' values
/// <summary>
/// <param name="index">The index of the current site</param>
/// <returns>The sum as an integer</returns>
int getNeighborSum(int index)
{
	int x = thisNode.sites[index].x, y = thisNode.sites[index].y, z = thisNode.sites[index].z;
	int top, bottom, left, right, up, down;
	
	/* get left and right neighbor values - use ghostsites if necessary */
	left = x == 0 ? thisNode.ghostSites[LEFT][z * thisNode.sites_on_ny + y].value : thisNode.sites[index - 1].value;	
    right = x == thisNode.sites_on_nx - 1 ? thisNode.ghostSites[RIGHT][z * thisNode.sites_on_ny + y].value : thisNode.sites[index + 1].value;
    
	/* get top and bottom neighbor values - use ghostsites if necessary */
    top = y == 0 ? thisNode.ghostSites[TOP][z * thisNode.sites_on_nx + x].value : thisNode.sites[index - thisNode.sites_on_nx].value;
    bottom = y == thisNode.sites_on_ny - 1 ? thisNode.ghostSites[BOTTOM][z * thisNode.sites_on_nx + x].value : thisNode.sites[index + thisNode.sites_on_nx].value;

	/* get up and down neighbor values - use ghostsites if necessary */
    up = z == 0 ? thisNode.ghostSites[UP][y * thisNode.sites_on_nx + x].value : thisNode.sites[index - thisNode.sites_on_nx * thisNode.sites_on_ny].value;
    down = z == thisNode.sites_on_nz - 1 ? thisNode.ghostSites[DOWN][y * thisNode.sites_on_nx + x].value : thisNode.sites[index + thisNode.sites_on_nx * thisNode.sites_on_ny].value;
	
	/* possible debug test to try ghost sites */
	/*left = x == 0 ? thisNode.sites[index + thisNode.sites_on_nx - 1].value : thisNode.sites[index - 1].value;
	right = x == thisNode.sites_on_nx - 1 ? thisNode.sites[index - thisNode.sites_on_nx + 1].value : thisNode.sites[index + 1].value;
    top = y == 0 ? thisNode.sites[index + thisNode.sites_on_nx * thisNode.sites_on_ny - thisNode.sites_on_nx].value : thisNode.sites[index - thisNode.sites_on_nx].value;
    bottom = y == thisNode.sites_on_ny - 1 ? thisNode.sites[index - thisNode.sites_on_nx * thisNode.sites_on_ny + thisNode.sites_on_nx].value : thisNode.sites[index + thisNode.sites_on_nx].value;
    up = 0;
    down = 0;*/

	return top + bottom + left + right + up + down;
}
